import xml.etree.ElementTree as ET
import pprint
import csv

'''
    When parsing an SEC Filing that contains XBLR files, it's curucial that you know how the files are structured
    and what each file contains. If you skip this portion it's very easy to get lost in the process and lose sight
    of how to get the data.

        CALCULATION FILE: TradingSymbol_EndDate_cal.xml
        This contains links to different calculations in the file this will be useful when trying to 
        reference how certain calculations were made.

        DEFINITION FILE: TradingSymbol_EndDate_def.xml
        This contains links to different definitions used in the 10K/Q and also external resources 
        that define those definitions, this would be entities like FASB.

        LABEL FILE: TradingSymbol_EndDate_lab.xml
        This contains a section for each label in the document along with their corresponding ID used 
        in the XML structuring. Additionally each one these labels also has a locator that will provide 
        a hyperlink to that label in the file.

        PRESENTATION FILE: TradingSymbol_EndDate_pre.xml
        This provides more links to different presentations that contain information regarding different 
        sections of the filing.

    The general strategy is as follows:

        1. Parse the labels file.
        2. Parse the definitions file, and match the information to the labels file.
        3. Parse the calculations file, and match the information to the labels file.
        4. Parse the maind HTML file.
'''


# Define the file path.
file_htm = r'C:\Users\Alex\OneDrive\Growth - Tutorial Videos\Lessons - SEC\0001326801-19-000069-xbrl\fb-09302019x10q.htm'
file_cal = r'C:\Users\Alex\OneDrive\Growth - Tutorial Videos\Lessons - SEC\0001326801-19-000069-xbrl\fb-20190930_cal.xml'
file_lab = r'C:\Users\Alex\OneDrive\Growth - Tutorial Videos\Lessons - SEC\0001326801-19-000069-xbrl\fb-20190930_lab.xml'
file_def = r'C:\Users\Alex\OneDrive\Growth - Tutorial Videos\Lessons - SEC\0001326801-19-000069-xbrl\fb-20190930_def.xml'

# # Initalize my master dictionary list, this will store the content as I parse it.
# dict_list = []

# '''
#     STEP 1: PARSE THE LABELS FILE.
# '''

# # Load the file into the `ElementTree.Parse` function.
# tree = ET.parse(file_lab)

# # Every `ElementTree` has a Root Node. This is the starting point of our tree.
# root_node = tree.getroot().tag

# # I use the root node, extract the namespace. The namespace will exist in every node we have, this can make parsing a little challenging.
# # I rather not have it at all and just remove it from the content as I'm parsing it.
# root_node_name_space =  root_node.split('}')[0] + '}'

# # Labels come in two forms, those I want and those I don't want.
# avoids = ['linkbase','roleRef']
# parse = ['label','labelLink','labelArc','loc','definitionLink','definitionArc','calculationArc']


# # Grab all the definition links.
# label_links = tree.findall(r'{http://www.xbrl.org/2003/linkbase}labelLink')

# # loop through each label in the tree.
# for label_link in label_links:
#     for label_element in label_link.iter():

#         # split the label to remove the namespace component, we now have a list.
#         split_label = label_element.tag.split('}')

#         # The first element is the namespace, and the second element is a label.
#         namespace = split_label[0]
#         label = split_label[1]

#         # Grab the label value and check if it's in the parse list.
#         if label in parse:

#             # create a dictionary for to store the values
#             dict_storage = {}
#             dict_storage['item_type'] = 'label_' + label

#             # grab the attribute keys
#             label_keys = label_element.keys()

#             # for each key.
#             for key in label_keys:
                
#                 # parse if needed.
#                 if '}' in key:

#                     # add the new key to the dictionary and grab the old value.
#                     new_key = key.split('}')[1]
#                     dict_storage[new_key] = link_label.attrib[key]
#                 else:
#                     # grabe the value.
#                     dict_storage[key] = link_label.attrib[key]

#             # add to the dictionary
#             dict_list.append(dict_storage)

# '''
#     STEP 2: PARSE THE DEFINITIONS FILE.
# '''

# # Parse the tree.
# tree = ET.parse(file_def)

# # Grab all the definition links.
# definition_links = tree.findall(r'{http://www.xbrl.org/2003/linkbase}definitionLink')

# # Loop through the definition links
# for definition_link in definition_links:

#     # each definition link can have it's own children, so loop through those.
#     for definition in definition_link.iter():

#         # split the label to remove the namespace component, we now have a list.
#         split_label = definition.tag.split('}')

#         # The first element is the namespace, and the second element is a label.
#         namespace = split_label[0]
#         label = split_label[1]

#         if label in parse:

#             # create a dictionary for to store the values
#             dict_storage = {}
#             dict_storage['item_type'] = 'definition_' + split_label[1]

#             # grab the attribute keys
#             def_keys = definition.keys()

#             # for each key.
#             for key in def_keys:

#                  # parse if needed.
#                 if '}' in key:

#                     # add the new key to the dictionary and grab the old value.
#                     new_key = key.split('}')[1]
#                     dict_storage[new_key] = definition.attrib[key]

#                 else:
#                     # grabe the value.
#                     dict_storage[key] = definition.attrib[key]

#             # add to dictionary.
#             dict_list.append(dict_storage)

# '''
#     STEP 3: PARSE THE CALCULATIONS FILE.
# '''

# # Parse the tree.
# tree = ET.parse(file_cal)

# # Grab all the calculation links.
# calculation_links = tree.findall(r'{http://www.xbrl.org/2003/linkbase}calculationLink')

# # Loop throught the calculation links
# for calculation_link in calculation_links:

#     # each calculation link can have their own children, so loop through those.
#     for calculation in calculation_link.iter():

#         # split the label to remove the namespace component, we now have a list.
#         split_label = calculation_link.tag.split('}')

#         # The first element is the namespace, and the second element is a label.
#         namespace = split_label[0]
#         label = split_label[1]

#         if label in parse:

#             dict_storage = {}
#             dict_storage['item_type'] = 'calculation_' + split_label[1]

#             # grab the attribute keys
#             cal_keys = calculation.keys()

#             # for each key.
#             for key in cal_keys:

#                  # parse if needed.
#                 if '}' in key:

#                     # add the new key to the dictionary and grab the old value.
#                     new_key = key.split('}')[1]
#                     dict_storage[new_key] = calculation.attrib[key]

#                 else:
#                     # grabe the value.
#                     dict_storage[key] = calculation.attrib[key]

#             # add to dictionary.
#             dict_list.append(dict_storage)

# Initalize some dictionaries to store some values.
storage_values = {}
storage_gaap = {}

# Initalize my master dictionary list, this will store the content as I parse it.
dict_list = []

# Initalize my list of files I plan to parse.
files_list = [
    (file_cal, r'{http://www.xbrl.org/2003/linkbase}calculationLink', 'calculation'), 
    (file_def, r'{http://www.xbrl.org/2003/linkbase}definitionLink','definition'), 
    (file_lab, r'{http://www.xbrl.org/2003/linkbase}labelLink','label')
    ]

# Labels come in two forms, those I want and those I don't want.
avoids = ['linkbase','roleRef']
parse = ['label','labelLink','labelArc','loc','definitionLink','definitionArc','calculationArc']

# part of the process is matching up keys, to do that we will store some keys as we parse them.
lab_list = set()
cal_list = set()

# loop through each file.
for file in files_list:

    # Parse the tree by passing through the file.
    tree = ET.parse(file[0])

    # Grab all the calculation links.
    links = tree.findall(file[1])

    # Loop throught the calculation links
    for link in links:

        # each calculation link can have their own children, so loop through those.
        for link_element in link.iter():

            # split the label to remove the namespace component, we now have a list.
            split_label = link_element.tag.split('}')

            # The first element is the namespace, and the second element is a label.
            namespace = split_label[0]
            label = split_label[1]

            # if it's a label we want then continue.
            if label in parse:

                # define the item type label
                item_type_label = file[2] + '_' + split_label[1]
                
                # initalize a smaller dictionary that will house all the content from that element.
                dict_storage = {}
                dict_storage['item_type'] = item_type_label

                # grab the attribute keys
                cal_keys = link_element.keys()

                # for each key.
                for key in cal_keys:

                    # parse if needed.
                    if '}' in key:

                        # add the new key to the dictionary and grab the old value.
                        new_key = key.split('}')[1]
                        dict_storage[new_key] = link_element.attrib[key]

                    else:
                        # grabe the value.
                        dict_storage[key] = link_element.attrib[key]

                # add important values to the set list.
                if item_type_label == 'label_label':

                    key_store = dict_storage['label']
                    master_key = key_store.replace('lab_','')

                    label_split = master_key.split('_')
                    label_split_joined = label_split[0] + ':' + label_split[1]
                    storage_values[master_key] = {}
                    storage_values[master_key][label_split[0]] = label_split_joined
                    storage_values[master_key]['label_id'] = key_store
                    storage_values[master_key]['location_id'] = key_store.replace('lab_','loc_')
                    storage_values[master_key]['us_gaap_id'] = label_split_joined
                    storage_values[master_key]['us_gaap_value'] = None
                    storage_values[master_key][item_type_label] = dict_storage 

                    storage_gaap[label_split_joined] = {}
                    storage_gaap[label_split_joined]['id'] = label_split_joined
                    storage_gaap[label_split_joined]['master_id'] = master_key


                # add to dictionary.
                dict_list.append([file[2], dict_storage])

'''
    PARSE THE HTML FILE.
'''

# Load the HTML file.
tree = ET.parse(file_htm)

# create a new dictionary to store context info.
context_dictionary = {}

# loop through all the elements in the HTML file.
for element in list(tree.iter()):
    
    # for nonNumber the content is different.
    if 'nonNumeric' in element.tag:
        storage_gaap[element.attrib['name']]['context_ref'] = element.attrib['contextRef']
        storage_gaap[element.attrib['name']]['context_id'] = element.attrib['id']
        storage_gaap[element.attrib['name']]['continued_at'] = element.attrib.get('continuedAt','null')
        storage_gaap[element.attrib['name']]['escape'] = element.attrib.get('escape','null')
        storage_gaap[element.attrib['name']]['escape'] = element.attrib.get('format','null')
        if storage_gaap[element.attrib['name']]['master_id'] in storage_values:
            storage_values[storage_gaap[element.attrib['name']]['master_id']]['us_gaap_value'] = storage_gaap[element.attrib['name']]  

    # same for nonFraction tags.
    if 'nonFraction' in element.tag:
        storage_gaap[element.attrib['name']]['context_ref'] = element.attrib['contextRef']
        storage_gaap[element.attrib['name']]['fraction_id'] = element.attrib['id']
        storage_gaap[element.attrib['name']]['unit_ref'] = element.attrib.get('unitRef','null')
        storage_gaap[element.attrib['name']]['decimals'] = element.attrib.get('decimals','null')
        storage_gaap[element.attrib['name']]['scale'] = element.attrib.get('scale','null')
        storage_gaap[element.attrib['name']]['format'] = element.attrib.get('format','null')
        storage_gaap[element.attrib['name']]['value'] = element.text.strip() if element.text else 'Null'
        if storage_gaap[element.attrib['name']]['master_id'] in storage_values:
            storage_values[storage_gaap[element.attrib['name']]['master_id']]['us_gaap_value'] = storage_gaap[element.attrib['name']]    

    
    # context is very different.
    if 'context' in element.tag:
        context_dictionary[element.attrib['id']] = {}
        for cnx_item in element.iter():
            for att in cnx_item.attrib:
                if att:
                    context_dictionary[element.attrib['id']][att] = cnx_item.attrib[att]
                    if cnx_item.text.strip() != '':
                        context_dictionary[element.attrib['id']]['text'] = cnx_item.text.strip()
                    else:
                        context_dictionary[element.attrib['id']]['text'] = 'Null'
    

# pprint.pprint(storage_values.keys())
# pprint.pprint(dict_list)
# for key in storage_gaap:
#     if storage_gaap[key]['master_id'] in storage_values:
#         storage_values[storage_gaap[key]['master_id']]['us_gaap_value'] = storage_gaap[key]



# pprint.pprint(context_dictionary)              

# INTERSTING DIRECTORIES - SEC
# country
# naics
# invest
# exch
# dei
# stpr
# sic
# currency

# INTERSTING DIRECTORIES - FASB
# us-gaap

# https://xbrl.sec.gov/invest/
# https://xbrl.sec.gov/sic/2020
# https://xbrl.sec.gov/country/2020/country-lab-2020-01-31.xml
# https://xbrl.sec.gov/naics/2017/naics-doc-2017-01-31.xml
# https://xbrl.sec.gov/currency/2020/currency-lab-2020-01-31.xml
# https://xbrl.sec.gov/exch/2020/
# https://xbrl.sec.gov/dei
# https://xbrl.sec.gov/stpr
# http://xbrl.fasb.org/us-gaap/

# https://www.sec.gov/data.json



# first write the xbrl_content.
file_name = 'sec_xbrl_content.csv'

# open the file.
with open(file_name, mode='w', newline='') as sec_file:

    # create the writer.
    writer = csv.writer(sec_file)

    # write the header.
    writer.writerow(['FILE','LABEL','VALUE'])

    # dump the dict to the csv file.
    for dict_cont in dict_list:
        for item in dict_cont[1].items():
            writer.writerow([dict_cont[0]] + list(item))


# second write the filing_values.
file_name = 'sec_xbrl_values.csv'

# open the file.
with open(file_name, mode='w', newline='') as sec_file:

    # create the writer.
    writer = csv.writer(sec_file)

    # write the header.
    writer.writerow(['ID','CATEGORY','LABEL','VALUE'])

    # start at level 1
    for storage_1 in storage_values:

        # level two is grab the items.
        for storage_2 in storage_values[storage_1].items():

            # if the value is a dictionary, we have one more possible level.
            if isinstance(storage_2[1], dict):

                # level three grab the items.
                for storage_3 in storage_2[1].items():

                    # Write the values to the csv.
                    writer.writerow([storage_1] + [storage_2[0]] + list(storage_3))

            # else just write it to the CSV.
            else:
                if storage_2[1] != None:
                    writer.writerow([storage_1] + list(storage_2) + ['None'])  


import requests
import xml.etree.ElementTree as ET

from typing import List
from typing import Union
from datetime import date
from datetime import datetime

from pysec.parser import EDGARParser

# https://www.sec.gov/cgi-bin/srch-edgar?text=form-type%3D%2810-q*+OR+10-k*%29&first=2020&last=2020


class EDGARQuery():

    def __init__(self):
        """Initalizes the EDGAR Client with the different endpoints used."""

        # base URL for the SEC EDGAR browser
        self.sec_url = "https://www.sec.gov"

        self.archive_service = "https://www.sec.gov/Archives/edgar"
        self.browse_service = "https://www.sec.gov/cgi-bin/browse-edgar"
        self.issuer_service = "https://www.sec.gov/cgi-bin/own-disp"
        self.search_service = "https://www.sec.gov/cgi-bin/srch-edgar"
        self.series_service = "https://www.sec.gov/cgi-bin/series"
        self.current_service = "https://www.sec.gov/cgi-bin/current"

        self.sec_cgi_endpoint = "https://www.sec.gov/cgi-bin"
        self.cik_lookup = 'cik_lookup'
        self.mutal_fund_search = 'series'

        self.parser_client = EDGARParser()

    def get_sec_datasets(self) -> dict:
        """Grabs all the Public datasets provided by the SEC.

        Returns:
        ----
        dict: A collection of SEC datasets.

        Usage:
        ----
            >>> edgar_client = EDGARQuery()
            >>> sec_datasets = edgar_client.get_sec_datasets()
            {
                "@context": "https://project-open-data.cio.gov/v1.1/schema/catalog.jsonld",
                "@id": "https://www.sec.gov/data.json",
                "@type": "dcat:Catalog",
                "conformsTo": "https://project-open-data.cio.gov/v1.1/schema",
                "describedBy": "https://project-open-data.cio.gov/v1.1/schema/catalog.json",
                "dataset": []
            }
        """

        # Make the request.
        response = requests.get(
            url='https://www.sec.gov/data.json'
        )

        if response.ok:
            return response.json()

    def get_edgar_taxonomies(self) -> dict:
        """Grabs all the Public taxonomies datasets provided by the SEC.

        Returns:
        ----
        dict: A collection of Taxonomy files for the SEC.

        Usage:
        ----
            >>> edgar_client = EDGARQuery()
            >>> sec_datasets = edgar_client.get_edgar_taxonomies()
            [
                {
                    'AttType': 'SCH',
                    'Elements': '0',
                    'Family': 'BASE',
                    'FileTypeName': 'Schema',
                    'Href': 'http://www.xbrl.org/2003/xbrl-linkbase-2003-12-31.xsd',
                    'Namespace': 'http://www.xbrl.org/2003/linkbase',
                    'Prefix': 'link',
                    'Version': '2010'
                },
                {
                    'AttType': 'SCH',
                    'Elements': '0',
                    'Family': 'BASE',
                    'FileTypeName': 'Schema',
                    'Href': 'http://www.xbrl.org/2003/xbrl-instance-2003-12-31.xsd',
                    'Namespace': 'http://www.xbrl.org/2003/instance',
                    'Prefix': 'xbrli',
                    'Version': '2010'
                }
            ]
        """

        # Make the request.
        response = requests.get(
            url='https://www.sec.gov/info/edgar/edgartaxonomies.xml'
        )

        # Parse the response.
        taxonomies = self.parser_client.parse_loc_elements(
            response_text=response.text
        )

        return taxonomies

    def company_directories(self, cik: str) -> dict:
        """Grabs all the filing directories for a company.

        Overview:
        ----
        Companies often file many SEC disclosures, so this endpoint
        makes grabbing all the endpoints associated with a company
        easy, by only requiring the CIK number.

        Arguments:
        ----
        cik {str} -- The company CIK number, defined by the SEC.

        Returns:
        ----
        dict -- A Dictionary containing the directory filings path.

        Usage:
        ----
            >>> edgar_client = EDGARQuery()
            >>> company_filings = edgar_client.company_directories(cik='1265107')
            [
                {
                    'last-modified': '2019-07-02 12:27:42',
                    'name': '000000000019010655',
                    'size': '',
                    'type': 'folder.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000000000019010655/index.json'
                },
                {
                    'last-modified': '2019-07-01 17:17:26',
                    'name': '000110465919038688',
                    'size': '',
                    'type': 'folder.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/index.json'
                }
            ]
        """

        # Build the URL.
        url = self.archive_service + "/data/{cik_number}/index.json".format(
            cik_number=cik
        )

        cleaned_directories = []
        directories = requests.get(url=url).json()

        # Loop through each item.
        for directory in directories['directory']['item']:

            # Create the URL.
            directory['url'] = self.archive_service + "/data/{cik_number}/{directory_id}/index.json".format(
                cik_number=cik,
                directory_id=directory['name']
            )

            directory['filing_id'] = directory.pop('name')
            directory['last_modified'] = directory.pop('last-modified')

            cleaned_directories.append(directory)

        return cleaned_directories

    def company_directory(self, cik: str, filing_id: str) -> dict:
        """Grabs all the items from a specific filing.

        Overview:
        ----
        The SEC organizes filings by CIK number which represent a single
        entity. Each entity can have multiple filings, which is identified
        by a filing ID. That filing can contain multiple items in it.

        This endpoint will return all the items from a specific filing that
        belongs to a single company.

        Arguments:
        ----
        cik {str} -- The company CIK number, defined by the SEC.

        filing_id {str} -- The ID of filing to pull.

        Returns:
        ----
        dict -- A Dictionary containing the filing items.

        Usage:
        ----
            >>> edgar_client = EDGARQuery()
            >>> company_filings = edgar_client.company_directory(cik='1265107', filing_id='000110465919038688')
            [
                {
                    'item_id': '0001104659-19-038688.txt',
                    'last_modified': '2019-07-01 17:17:26',
                    'size': '',
                    'type': 'text.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/0001104659-19-038688.txt'
                },
                {
                    'item_id': 'a19-12321_2425.htm',
                    'last_modified': '2019-07-01 17:17:26',
                    'size': '37553',
                    'type': 'text.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/a19-12321_2425.htm'
                }
            ]
        """

        url = self.archive_service + "/data/{cik_number}/{filing_number}/index.json".format(
            cik_number=cik,
            filing_number=filing_id
        )

        cleaned_items = []
        directory = requests.get(url=url).json()

        for item in directory['directory']['item']:

            item['url'] = self.archive_service + "/data/{cik_number}/{directory_id}/{file_id}".format(
                cik_number=cik,
                directory_id=filing_id,
                file_id=item['name']
            )

            item['item_id'] = item.pop('name')
            item['last_modified'] = item.pop('last-modified')
            cleaned_items.append(item)

        return cleaned_items

    def company_filings(self, cik: str = None, filing_type: str = None, sic_code: str = None, filing_number: str = None, company_name: str = None,
                        state: str = None, country: str = None, return_count: int = 100, start: int = 0, before: Union[str, date] = None,
                        after: Union[str, date] = None) -> List[dict]:
        """Returns all the filings of certain type for a particular company.

        Arguments:
        ----
        cik {str} -- The company CIK Number.

        filing_type {str} -- The filing type ID.

        Returns:
        ----
        dict -- A Dictionary containing the filing items.

        Usage:
        ----
            >>> edgar_client = EDGARQuery()
            >>> company_filings = edgar_client.company_directory(cik='1265107', filing_id='000110465919038688')
            [
                {
                    'item_id': '0001104659-19-038688.txt',
                    'last_modified': '2019-07-01 17:17:26',
                    'size': '',
                    'type': 'text.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/0001104659-19-038688.txt'
                },
                {
                    'item_id': 'a19-12321_2425.htm',
                    'last_modified': '2019-07-01 17:17:26',
                    'size': '37553',
                    'type': 'text.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/a19-12321_2425.htm'
                }
            ]
        """

        # Set the params
        params = {
            'action': 'getcompany',
            'output': 'atom',
            'sic': sic_code,
            'CIK': cik,
            'type': filing_type,
            'filenum': filing_number,
            'company': company_name,
            'start': start,
            'datea': after,
            'dateb': before
        }

        # Grab the response.
        response = requests.get(
            url=self.browse_service,
            params=params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count,
            start=start
        )

        return entries

    def company_filings_by_type(self, cik: str, filing_type: str) -> List[dict]:
        """Returns all the filings of certain type for a particular company.

        Arguments:
        ----
        cik {str} -- The company CIK Number.

        filing_type {str} -- The filing type ID.

        Returns:
        ----
        dict -- A Dictionary containing the filing items.

        Usage:
        ----
            >>> edgar_client = EDGARQuery()
            >>> company_filings = edgar_client.company_directory(cik='1265107', filing_id='000110465919038688')
            [
                {
                    'item_id': '0001104659-19-038688.txt',
                    'last_modified': '2019-07-01 17:17:26',
                    'size': '',
                    'type': 'text.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/0001104659-19-038688.txt'
                },
                {
                    'item_id': 'a19-12321_2425.htm',
                    'last_modified': '2019-07-01 17:17:26',
                    'size': '37553',
                    'type': 'text.gif',
                    'url': 'https://www.sec.gov/Archives/edgar/data/1265107/000110465919038688/a19-12321_2425.htm'
                }
            ]
        """

        # Set the params
        params = {
            'action': 'getcompany',
            'CIK': cik,
            'type': filing_type,
            'output': 'atom'
        }

        # Grab the response.
        response = requests.get(
            url=self.browse_service,
            params=params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def companies_by_state(self, state: str, return_count: int = 100) -> List[dict]:
        """Returns all the companies that fall under a given state.

        Arguments:
        ----
        state {str} -- [description]

        Returns:
        ----
        List[dict] -- [description]
        """

        # define the arguments of the request
        search_sic_params = {
            'State': state,
            'Count': '100',
            'action': 'getcompany',
            'output': 'atom'
        }

        response = requests.get(
            url=self.browse_service,
            params=search_sic_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count
        )

        return entries

    def companies_by_country(self, country: str, return_count: int = 100) -> List[dict]:
        """Grabs all the companies that fall under a particular country code.

        Arguments:
        ----
        country {str} -- The country code.

        Keyword Arguments:
        ----
        return_count {int} -- If you would like to limit the number of results, then
            specify the number of companies you want back. (default: {None})

        Returns:
        ----
        List[dict] -- A list of Entry resources.
        """

        # define the arguments of the request
        search_sic_params = {
            'Country': country,
            'Count': '100',
            'action': 'getcompany',
            'output': 'atom'
        }

        # Grab the Response.
        response = requests.get(
            url=self.browse_service,
            params=search_sic_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count
        )

        return entries

    def companies(self, sic_code: str = None, state: str = None, country: str = None, return_count: int = 100, start: int = 0) -> List[dict]:
        """Used to grab all companies matching multiple criteria.

        Overview:
        ----
        Returns all companies, that fall under a particular filter. This 
        endpoint allows you to grab companies matching multiple criteria 
        instead of a single criteria like in the other endpoints

        Arguments:
        ----
        sic_code {str} -- The SIC code for a particular Industry.

        state {str} -- The two character State Code.

        country {str} -- The two character country code.

        return_count {int} -- If you would like to limit the number of results, then
            specify the number of companies you want back. (default: {None})

        start {int} -- Specifies the starting company number. (default: {None})

        Returns:
        ----
            list[dict] -- A list of companies with the following attributes:

            [
                {
                    "state": "MN",
                    "cik": "0000066740",
                    "last-date": "",
                    "name": "3M CO",
                    "sic-code": "3841",
                    "id": "urn:tag:www.sec.gov:cik=0000066740",
                    "href": "URL",
                    "type": "html",
                    "summary": "<strong>CIK:</strong> 0000066740, <strong>State:</strong> MN",
                    "title": "3M CO",
                    "updated": "2020-04-05T15:21:24-04:00",
                    "atom_owner_only": "URL",
                    "atom_owner_exclude": "URL",
                    "atom_owner_include": "URL",
                    "html_owner_only": "URL",
                    "html_owner_exclude": "URL",
                    "html_owner_include": "URL",
                    "atom_owner_only_filtered_date": "URL",
                    "atom_owner_exclude_filtered_date": "URL",
                    "atom_owner_include_filtered_date": "URL",
                    "html_owner_only_filtered_date": "URL",
                    "html_owner_exclude_filtered_date": "URL",
                    "html_owner_include_filtered_date": "URL",
                }
            ]
        """

        if not start:
            start = 0

        # define the arguments of the request
        search_sic_params = {
            'Count': '100',
            'SIC': sic_code,
            'Country': country,
            'State': state,
            'action': 'getcompany',
            'output': 'atom',
            'start': start
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_sic_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count,
            start=start
        )

        return entries

    def companies_by_sic(self, sic_code: str, return_count: int = 100, start: int = 0) -> List[dict]:
        """Grabs all companies with a certain SIC code.

        Returns all companies, that fall under a particular SIC code. The information returned
        by this endpoint depends on the infromation available on the company.

        Arguments:
        ----
        sic_code {str} -- The SIC code for a particular Industry.

        Keyword Arguments:
        ----
        return_count {int} -- If you would like to limit the number of results, then
            specify the number of companies you want back. (default: {None})

        start {int} -- Specifies the starting company number. (default: {None})

        Returns:
        ----
            list[dict] -- A list of companies with the following attributes:

            [
                {
                    "state": "MN",
                    "cik": "0000066740",
                    "last-date": "",
                    "name": "3M CO",
                    "sic-code": "3841",
                    "id": "urn:tag:www.sec.gov:cik=0000066740",
                    "href": "URL",
                    "type": "html",
                    "summary": "<strong>CIK:</strong> 0000066740, <strong>State:</strong> MN",
                    "title": "3M CO",
                    "updated": "2020-04-05T15:21:24-04:00",
                    "atom_owner_only": "URL",
                    "atom_owner_exclude": "URL",
                    "atom_owner_include": "URL",
                    "html_owner_only": "URL",
                    "html_owner_exclude": "URL",
                    "html_owner_include": "URL",
                    "atom_owner_only_filtered_date": "URL",
                    "atom_owner_exclude_filtered_date": "URL",
                    "atom_owner_include_filtered_date": "URL",
                    "html_owner_only_filtered_date": "URL",
                    "html_owner_exclude_filtered_date": "URL",
                    "html_owner_include_filtered_date": "URL",
                }
            ]
        """

        if not start:
            start = 0

        # define the arguments of the request
        search_sic_params = {
            'SIC': sic_code,
            'Count': '100',
            'action': 'getcompany',
            'output': 'atom',
            'start': start
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_sic_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count,
            start=start
        )

        return entries

    # OWNERSHIP FILINGS

    def ownership_filings(self, cik: str = None, company_name: str = None, start: int = 0,
                          return_count: int = 100, before: Union[str, date] = None, after: Union[str, date] = None) -> List[dict]:
        """Returns all the ownership filings for the criteria specified.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        return_count {int} -- If you would like to limit the number of results, then
            specify the number of companies you want back. (default: {None})

        start {int} -- Specifies the starting company number. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'company': company_name,
            'Count': '100',
            'myowner': 'only',
            'action': 'getcompany',
            'output': 'atom',
            'start': start,
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count,
            start=start
        )

        return entries

    def ownership_filings_by_cik(self, cik: str, before: Union[str, date] = None, after: Union[str, date] = None) -> List[dict]:
        """Returns all the ownership filings for a given CIK number in a given date range.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'only',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def ownership_filings_by_company_name(self, company_name: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the filings ownership for a given company in a given date range.

        Arguments:
        ----
        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'company': company_name,
            'Count': '100',
            'myowner': 'only',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def non_ownership_filings(self, cik: str = None, company_name: str = None, start: int = 0, return_count: int = 100,
                              before: Union[str, date] = None, after: Union[str, date] = None) -> List[dict]:
        """Returns all the non-ownership filings for the criteria specified.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        return_count {int} -- If you would like to limit the number of results, then
            specify the number of companies you want back. (default: {None})

        start {int} -- Specifies the starting company number. (default: {None})
        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'exclude',
            'action': 'getcompany',
            'output': 'atom',
            'company': company_name,
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count,
            start=start
        )

        return entries

    def non_ownership_filings_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the non-ownership filings for a given CIK number in a given date range.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'exclude',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def non_ownership_filings_by_company_name(self, company_name: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the filings non-ownership for a given company in a given date range.

        Arguments:
        ----
        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'company': company_name,
            'Count': '100',
            'myowner': 'exclude',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def all_filings_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the filings (ownership and non-ownership) for a given CIK number in a given date range.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def all_filings_by_company_name(self, company_name: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the filings (ownership and non-ownership) for a given company in a given date range.

        Arguments:
        ----
        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of ownership filings.
        """

        # define the arguments of the request
        search_params = {
            'company': company_name,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_issuers_by_cik(self, cik: str) -> List[dict]:
        """Returns all the issuers for a given CIK number.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Returns:
        ----
        List[dict] -- A list of Issuer documents.
        """

        # define the arguments of the request
        search_params = {
            'count': '100',
            'CIK': cik,
            'action': 'getissuer'
        }

        # Make the response.
        response = requests.get(
            url=self.issuer_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_issuer_table(
            entries_text=response.text
        )

        return entries

    def mutual_funds(self, mutual_fund_type: str = None, cik: str = None, company_name: str = None, start: int = 0, return_count: int = 100,
                        before: str = None, after: str = None) -> List[dict]:
        """Returns all mutual funds for a given name in a given date range.

        Arguments:
        ----
        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual funds.
        """

        # The common mutual fund types.
        mutual_funds_type = {
            'mutual-fund': {
                'query_params': {
                    'type': '485',
                    'hidefilings': None
                }
            },
            'mutual-fund-prospectus': {
                'query_params': {
                    'type': '485',
                    'hidefilings': '0'
                }
            },
            'mutual-fund-proxy-records': {
                'query_params': {
                    'type': 'N-PX',
                    'hidefilings': '0'
                }
            },
            'mutual-fund-shareholder-reports': {
                'query_params': {
                    'type': 'N-CSR',
                    'hidefilings': '0'
                }
            },
            'mutual-fund-summary-prospectus': {
                'query_params': {
                    'type': '497K',
                    'hidefilings': '0'
                }
            },
            'mutual-fund-effectiveness-notice': {
                'query_params': {
                    'type': 'EFFECT',
                    'hidefilings': None
                }
            }
        }
        
        # Grab the one specified.
        query_params = mutual_funds_type[mutual_fund_type]

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'company': company_name,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Add the additional arguments.
        search_params.update(query_params)

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(
            entries_text=response.text,
            num_of_items=return_count,
            start=start
        )

        return entries

    def get_mutual_funds_by_name(self, company_name: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all mutual funds for a given name in a given date range.

        Arguments:
        ----
        company_name {str} -- The name of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual funds.
        """

        # define the arguments of the request
        search_params = {
            'company': company_name,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': '485',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_mutual_funds_prospectus_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the filings (ownership and non-ownership) for a given company in a given date range.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': '485',
            'hidefilings': '0',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_mutual_funds_proxy_records_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the Mutual Fund Proxy records for a given CIK number.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': 'N-PX',
            'hidefilings': '0',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_mutual_funds_shareholder_reports_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the Mutual Fund Proxy Shareholder Reports for a given CIK number.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': 'N-CSR',
            'hidefilings': '0',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_mutual_funds_statutory_prospectus_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the Mutual Fund Statutory Prospectus for a given CIK number.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': '485',
            'hidefilings': '0',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_mutual_funds_summary_prospectus_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the Mutual Fund Summary Prospectus for a given CIK number.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': '497K',
            'hidefilings': '0',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_mutual_funds_effectiveness_notices_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the Mutual Fund Summary Prospectus for a given CIK number.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': 'EFFECT',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.browse_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_variable_insurance_products_by_name(self, product_name: str) -> List[dict]:
        """Returns all the variable insurance products defined by the given name.

        Arguments:
        ----
        product_name {str} -- Variable insurance products.

        Returns:
        ----
        List[dict] -- A list of mutual funds.
        """

        # define the arguments of the request
        search_params = {
            'company': product_name,
            'sc': 'companyseries'
        }

        # Make the response.
        response = requests.get(
            url=self.series_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_variable_products_company_table(
            product_table_page=response.text
        )

        return entries

    def get_variable_insurance_products_by_cik(self, cik: str, before: str = None, after: str = None) -> List[dict]:
        """Returns all the filings (ownership and non-ownership) for a given company in a given date range.

        Arguments:
        ----
        cik {str} -- The CIK number of the company to be queried.

        Keyword Arguments:
        ----
        before {Union[str, date]} -- Represents filings that you want before a certain
            date. For example, "2019-12-01" means return all the filings BEFORE
            Decemeber 1, 2019. (default: {None})

        after {Union[str, date]} -- Represents filings that you want after a certain
            date. For example, "2019-12-01" means return all the filings AFTER 
            Decemeber 1, 2019. (default: {None})

        Returns:
        ----
        List[dict] -- A list of mutual fund prospectus.
        """

        # define the arguments of the request
        search_params = {
            'CIK': cik,
            'Count': '100',
            'myowner': 'include',
            'action': 'getcompany',
            'type': '485',
            'hidefilings': '0',
            'output': 'atom',
            'datea': after,
            'dateb': before
        }

        # Make the response.
        response = requests.get(
            url=self.series_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_entries(entries_text=response.text)

        return entries

    def get_current_event_filings(self, days_prior: int, form: str, form_id: str = None) -> List[dict]:
        """Query current filigns by type.

        Arguments:
        ----
        day (int): The number of days prior you would like to analyze. Can be one of
            the following: [0, 1, 2, 3, 4, 5]

        form (str): The form you would like to analyze. Can be one of
            the following: ['10-k-annual', '10-k-quarterly', '14-proxies', '485-fund-prosp.', '8-k', 's-8', 'all']

        form_id (str, optional): Represents the Form-ID and can be used to override the `form` argument. Defaults to None.

        Returns:
        ----
        List[Dict]: A list of current event filings.
        """

        form_dict = {
            '10-k-annual': 0,
            '10-k-quarterly': 1,
            '14-proxies': 2,
            '485-fund-prosp.': 3,
            '8-k': 4,
            's-8': 5,
            'all': 6
        }

        if form in form_dict:
            form_id_arg = form_dict[form]
        else:
            raise ValueError(
                "The argument you've provided for form is incorrect.")

        # define the arguments of the request
        search_params = {
            'q1': days_prior,
            'q2': form_id_arg,
            'q3': form_id
        }

        # Make the response.
        response = requests.get(
            url=self.current_service,
            params=search_params
        )

        # Parse the entries.
        entries = self.parser_client.parse_current_event_table(
            current_event_page=response.text)

        return entries

    def get_quarterly_index(self, year: int, quarter: int) -> List[dict]:       
        """Grabs all the filings belong to a specific quarter and year.

        Arguments:
        ----
        year (int): The year from which to query. Must be >= 1994.

        quarter (int): The quarter from which to query. Can be one the
            following: [1, 2, 3, 4]

        Returns:
        ----
        List[dict]: A list of directory links and directory files.
        """

        if (year < 1994) or (year > datetime.now().year):
            raise ValueError("The year you specified can not be pulled.")

        if quarter not in [1, 2, 3, 4]:
            raise ValueError("The quarter you specified is incorrect.")

        quarterly_url = self.archive_service + "/full-index/{year}/{qtr}/".format(
            year=year,
            qtr="QTR" + str(quarter)
        )

        # Define the full URL.
        full_url = quarterly_url + "index.json"

        # Make the response.
        directories = requests.get(
            url=full_url
        ).json()

        cleaned_directories = []

        # Loop through each item.
        for item in directories['directory']['item']:
            
            new_item = {
                key.replace("-", "_"): 
                value for key, value in item.items()
            }

            new_item['url'] = quarterly_url + item['href']
            cleaned_directories.append(new_item)

        return cleaned_directories

    def get_quarterly_indexes(self) -> List[dict]:
        """Grabs all the Quarterly Index filings.

        Returns:
        ----
        List[dict]: A list of directory links and directory files.
        """

        url = self.archive_service + "/full-index/index.json"

        cleaned_directories = []

        # Make the response.
        directories = requests.get(
            url=url
        ).json()

        # Loop through each item.
        for directory in directories['directory']['item']:

            # Create the URL for the directory.
            if directory['type'] == 'dir':

                directory['yearly_url'] = self.archive_service + "/full-index/{year}/index.json".format(
                    year=directory['name'],
                )

                print('Pulling Directory: {yearly_url}'.format(
                    yearly_url=directory['yearly_url']))

                # If we have a year than grab the quarters.
                yearly_content = requests.get(
                    url=directory['yearly_url']).json()

                directory['quarterly_directories'] = {}

                for quarter in yearly_content['directory']['item']:

                    quarterly_url = self.archive_service + "/full-index/{year}/{qtr}/index.json".format(
                        year=directory['name'],
                        qtr=quarter['name']
                    )

                    # Define the URL with no JSON index.
                    quarterly_url_no_json = self.archive_service + "/full-index/{year}/{qtr}/".format(
                        year=directory['name'],
                        qtr=quarter['name']
                    )

                    print(
                        'Pulling Directory: {qtr_url}'.format(
                            qtr_url=quarterly_url
                        )
                    )

                    directory['quarterly_directories'][quarter['name']] = {}
                    directory['quarterly_directories'][quarter['name']]['quarterly_url'] = quarterly_url
                    directory['quarterly_directories'][quarter['name']]['items'] = []

                    quarterly_content = requests.get(url=quarterly_url).json()

                    # Loop through each file.
                    for item in quarterly_content['directory']['item']:
                        
                        # If 
                        is_gz = '.gz' in item['href']
                        is_idx = '.idx' in item['href']
                        is_zip = '.zip' in item['href']
                        is_Z = '.Z' in item['href']
                        is_xml = '.xml' in item['href']

                        if is_gz or is_idx or is_zip or is_Z or is_xml:
                            item['url'] = quarterly_url_no_json + item['href']
                        else:
                            item['url'] = quarterly_url + item['href']

                        directory['quarterly_directories'][quarter['name']]['items'].append(item)

            # Create the URL for downloads.
            else:

                directory['file_url'] = self.archive_service + "/full-index/{href}".format(
                    href=directory['href']
                )

            directory['filing_id'] = directory.pop('name')
            directory['last_modified'] = directory.pop('last-modified')

            cleaned_directories.append(directory)

        return cleaned_directories



from enum import Enum

class FILING_TYPES(Enum):

    _1012B='1012B'
    _1012BAD='1012BAD'
    _1012G='1012G'
    _1012GAD='1012GAD'
    _10C='10C'
    _10CAD='10CAD'
    _10D='10D'
    _10DAD='10DAD'
    _10K='10K'
    _10KAD='10KAD'
    _10K405='10K405'
    _10K405AD='10K405AD'
    _10KSB='10KSB'
    _10KSBAD='10KSBAD'
    _10KSB40='10KSB40'
    _10KSB40AD='10KSB40AD'
    _10KT='10KT'
    _10KTAD='10KTAD'
    _10KT405='10KT405'
    _10KT405AD='10KT405AD'
    _10Q='10Q'
    _10QAD='10QAD'
    _10QSB='10QSB'
    _10QSBAD='10QSBAD'
    _10QT='10QT'
    _10QTAD='10QTAD'
    _10SB12B='10SB12B'
    _10SB12BAD='10SB12BAD'
    _10SB12G='10SB12G'
    _10SB12GAD='10SB12GAD'
    _11K='11K'
    _11KAD='11KAD'
    _11KT='11KT'
    _11KTAD='11KTAD'
    _13FE='13FE'
    _13FEAD='13FEAD'
    _13FHR='13FHR'
    _13FHRAD='13FHRAD'
    _13FNT='13FNT'
    _13FNTAD='13FNTAD'
    _144='144'
    _144AD='144AD'
    _1512B='1512B'
    _1512BAD='1512BAD'
    _1512G='1512G'
    _1512GAD='1512GAD'
    _1515D='1515D'
    _1515DAD='1515DAD'
    _15F12B='15F12B'
    _15F12BAD='15F12BAD'
    _15F12G='15F12G'
    _15F12GAD='15F12GAD'
    _15F15D='15F15D'
    _15F15DAD='15F15DAD'
    _1812B='1812B'
    _1812BAD='1812BAD'
    _1812G='1812G'
    _1812GAD='1812GAD'
    _18K='18K'
    _18KAD='18KAD'
    _1A='1A'
    _1AAD='1AAD'
    _1APOS='1APOS'
    _1AW='1AW'
    _1AWAD='1AWAD'
    _1E='1E'
    _1EAD='1EAD'
    _1K='1K'
    _1KAD='1KAD'
    _1SA='1SA'
    _1SAAD='1SAAD'
    _1U='1U'
    _1UAD='1UAD'
    _1Z='1Z'
    _1ZAD='1ZAD'
    _1ZW='1ZW'
    _1ZWAD='1ZWAD'
    _20F='20F'
    _20FAD='20FAD'
    _20FR12B='20FR12B'
    _20FR12BAD='20FR12BAD'
    _20FR12G='20FR12G'
    _20FR12GAD='20FR12GAD'
    _24F2EL='24F2EL'
    _24F2ELAD='24F2ELAD'
    _24F2NT='24F2NT'
    _24F2NTAD='24F2NTAD'
    _24F2TM='24F2TM'
    _24F2TMAD='24F2TMAD'
    _24F3NT='24F3NT'
    _24F3NTAD='24F3NTAD'
    _25='25'
    _25AD='25AD'
    _253G1='253G1'
    _253G2='253G2'
    _253G3='253G3'
    _253G4='253G4'
    _25NSE='25NSE'
    _25NSEAD='25NSEAD'
    _2E='2E'
    _2EAD='2EAD'
    _3='3'
    _3AD='3AD'
    _305B2='305B2'
    _305B2AD='305B2AD'
    _35APP='35APP'
    _35APPAD='35APPAD'
    _35CERT='35CERT'
    _35CERTAD='35CERTAD'
    _4='4'
    _4AD='4AD'
    _4017F1='4017F1'
    _4017F1AD='4017F1AD'
    _4017F2='4017F2'
    _4017F2AD='4017F2AD'
    _4017G='4017G'
    _4017GAD='4017GAD'
    _4017GCS='4017GCS'
    _4017GCSAD='4017GCSAD'
    _4024B2='4024B2'
    _4024B2AD='4024B2AD'
    _4033='4033'
    _4033AD='4033AD'
    _406B='406B'
    _406BAD='406BAD'
    _408B25='408B25'
    _408F2='408F2'
    _408F2AD='408F2AD'
    _408FA='408FA'
    _408FAAD='408FAAD'
    _408FB='408FB'
    _408FBAD='408FBAD'
    _408FL='408FL'
    _408FLAD='408FLAD'
    _408FLA='408FLA'
    _408FLM='408FLM'
    _408FLMA='408FLMA'
    _408FM='408FM'
    _408FMAD='408FMAD'
    _40APP='40APP'
    _40APPAD='40APPAD'
    _40F='40F'
    _40FAD='40FAD'
    _40FR12B='40FR12B'
    _40FR12BAD='40FR12BAD'
    _40FR12G='40FR12G'
    _40FR12GAD='40FR12GAD'
    _40OIP='40OIP'
    _40OIPAD='40OIPAD'
    _424A='424A'
    _424B1='424B1'
    _424B2='424B2'
    _424B3='424B3'
    _424B4='424B4'
    _424B5='424B5'
    _424B7='424B7'
    _424B8='424B8'
    _424H='424H'
    _424HAD='424HAD'
    _425='425'
    _45B3='45B3'
    _45B3AD='45B3AD'
    _485A24E='485A24E'
    _485A24F='485A24F'
    _485APOS='485APOS'
    _485B24E='485B24E'
    _485B24F='485B24F'
    _485BPOS='485BPOS'
    _485BXT='485BXT'
    _485BXTE='485BXTE'
    _485BXTF='485BXTF'
    _486A24F='486A24F'
    _486APOS='486APOS'
    _486B24F='486B24F'
    _486BPOS='486BPOS'
    _487='487'
    _497='497'
    _497AD='497AD'
    _497H2='497H2'
    _497J='497J'
    _497K='497K'
    _497K1='497K1'
    _497K2='497K2'
    _497K3A='497K3A'
    _497K3B='497K3B'
    _5='5'
    _5AD='5AD'
    _6BNTC='6BNTC'
    _6BORDR='6BORDR'
    _6K='6K'
    _6KAD='6KAD'
    _8A12B='8A12B'
    _8A12BAD='8A12BAD'
    _8A12G='8A12G'
    _8A12GAD='8A12GAD'
    _8B12B='8B12B'
    _8B12BAD='8B12BAD'
    _8B12G='8B12G'
    _8F2NTC='8F2NTC'
    _8F2ORDR='8F2ORDR'
    _8K='8K'
    _8KAD='8KAD'
    _8K12B='8K12B'
    _8K12BAD='8K12BAD'
    _8K12G3='8K12G3'
    _8K12G3AD='8K12G3AD'
    _8K15D5='8K15D5'
    _8K15D5AD ='8K15D5AD'
    ABS15G='ABS15G'
    ABS15GAD='ABS15GAD'
    ABSEE='ABSEE'
    ABSEEAD='ABSEEAD'
    ANNLRPT='ANNLRPT'
    ANNLRPTAD='ANNLRPTAD'
    APPNTC='APPNTC'
    APPORDR='APPORDR'
    APPWD='APPWD'
    APPWDAD='APPWDAD'
    ARS='ARS'
    ARSAD='ARSAD'
    AW='AW'
    AWWD='AWWD'
    BW2='BW2'
    BW3='BW3'
    C='C'
    CAD='CAD'
    CAWD='CAWD'
    CAR='CAR'
    CARAD='CARAD'
    CARAWD='CARAWD'
    CARW='CARW'
    CB='CB'
    CBAD='CBAD'
    CFPORTAL='CFPORTAL'
    CFPORTALAD='CFPORTALAD'
    CFPORTALWD='CFPORTALWD'
    CORRESP='CORRESP'
    CTORDER='CTORDER'
    CTR='CTR'
    CTRAD='CTRAD'
    CTRW='CTRW'
    CU='CU'
    CUAD='CUAD'
    CUW='CUW'
    CW='CW'
    CWAD='CWAD'
    D='D'
    DAD='DAD'
    DEF13E3='DEF13E3'
    DEF13E3AD='DEF13E3AD'
    DEF14A='DEF14A'
    DEF14C='DEF14C'
    DEFA14A='DEFA14A'
    DEFA14C='DEFA14C'
    DEFC14A='DEFC14A'
    DEFC14C='DEFC14C'
    DEFM14A='DEFM14A'
    DEFM14C='DEFM14C'
    DEFN14A='DEFN14A'
    DEFR14A='DEFR14A'
    DEFR14C='DEFR14C'
    DEFS14A='DEFS14A'
    DEFS14C='DEFS14C'
    DELAM='DELAM'
    DFAN14A='DFAN14A'
    DFRN14A='DFRN14A'
    DOS='DOS'
    DOSA='DOSA'
    DOSLTR='DOSLTR'
    DRS='DRS'
    DRSAD='DRSAD'
    DRSLTR='DRSLTR'
    DSTRBRPT='DSTRBRPT'
    DSTRBRPTAD='DSTRBRPTAD'
    EFFECT='EFFECT'
    ETR='ETR'
    F1='F1'
    F1AD='F1AD'
    F10='F10'
    F10AD='F10AD'
    F10EF='F10EF'
    F10MEF='F10MEF'
    F10POS='F10POS'
    F1MEF='F1MEF'
    F2='F2'
    F2AD='F2AD'
    F2D='F2D'
    F2DPOS='F2DPOS'
    F2MEF='F2MEF'
    F3='F3'
    F3AD='F3AD'
    F3ASR='F3ASR'
    F3D='F3D'
    F3DPOS='F3DPOS'
    F3MEF='F3MEF'
    F4='F4'
    F4AD='F4AD'
    F4EF='F4EF'
    F4EFAD='F4EFAD'
    F4MEF='F4MEF'
    F4POS='F4POS'
    F6='F6'
    F6AD='F6AD'
    F6EF='F6EF'
    F6POS='F6POS'
    F7='F7'
    F7AD='F7AD'
    F7POS='F7POS'
    F8='F8'
    F8AD='F8AD'
    F80='F80'
    F80AD='F80AD'
    F80POS='F80POS'
    F8POS='F8POS'
    F9='F9'
    F9AD='F9AD'
    F9EF='F9EF'
    F9MEF='F9MEF'
    F9POS='F9POS'
    FN='FN'
    FNAD='FNAD'
    FWP='FWP'
    FX='FX'
    FXAD='FXAD'
    IRANNOTICE='IRANNOTICE'
    MA='MA'
    MAAD='MAAD'
    MAA='MAA'
    MAI='MAI'
    MAIAD='MAIAD'
    MANR='MANR'
    MAW='MAW'
    N1='N1'
    N1AD='N1AD'
    N14='N14'
    N14AD='N14AD'
    N148C='N148C'
    N148CAD='N148CAD'
    N14AE='N14AE'
    N14AEAD='N14AEAD'
    N14AE24='N14AE24'
    N14AE24AD='N14AE24AD'
    N14EL24='N14EL24'
    N14EL24AD='N14EL24AD'
    N14MEF='N14MEF'
    N18F1='N18F1'
    N18F1AD='N18F1AD'
    N1A='N1A'
    N1AAD='N1AAD'
    N1AEL='N1AEL'
    N1AELAD='N1AELAD'
    N2='N2'
    N2AD='N2AD'
    N23C1='N23C1'
    N23C1AD='N23C1AD'
    N23C2='N23C2'
    N23C2AD='N23C2AD'
    N23C3A='N23C3A'
    N23C3AAD='N23C3AAD'
    N23C3B='N23C3B'
    N23C3BAD='N23C3BAD'
    N23C3C='N23C3C'
    N23C3CAD='N23C3CAD'
    N27D1='N27D1'
    N27D1AD='N27D1AD'
    N2MEF='N2MEF'
    N3='N3'
    N3AD='N3AD'
    N30B2='N30B2'
    N30D='N30D'
    N30DAD='N30DAD'
    N3EL='N3EL'
    N3ELAD='N3ELAD'
    N4='N4'
    N4AD='N4AD'
    N4EL='N4EL'
    N4ELAD='N4ELAD'
    N5='N5'
    N5AD='N5AD'
    N54A='N54A'
    N54AAD='N54AAD'
    N54C='N54C'
    N54CAD='N54CAD'
    N6='N6'
    N6AD='N6AD'
    N6C9='N6C9'
    N6C9AD='N6C9AD'
    N6F='N6F'
    N6FAD='N6FAD'
    N8A='N8A'
    N8AAD='N8AAD'
    N8B2='N8B2'
    N8B2AD='N8B2AD'
    N8B3='N8B3'
    N8B3AD='N8B3AD'
    N8B4='N8B4'
    N8B4AD='N8B4AD'
    N8F='N8F'
    N8FAD='N8FAD'
    N8FNTC='N8FNTC'
    N8FORDR='N8FORDR'
    NCEN='NCEN'
    NCENDA='NCENDA'
    NCR='NCR'
    NCRAD='NCRAD'
    NCSR='NCSR'
    NCSRAD='NCSRAD'
    NCSRS='NCSRS'
    NCSRSAD='NCSRSAD'
    NMFP='NMFP'
    NMFPAD='NMFPAD'
    NMFP1='NMFP1'
    NMFP1AD='NMFP1AD'
    NMFP2='NMFP2'
    NMFP2AD='NMFP2AD'
    NPX='NPX'
    NPXAD='NPXAD'
    NQ='NQ'
    NQAD='NQAD'
    NRSROCE='NRSROCE'
    NRSROCEAD='NRSROCEAD'
    NRSROUPD='NRSROUPD'
    NRSROUPDAD='NRSROUPDAD'
    NSARA='NSARA'
    NSARAAD='NSARAAD'
    NSARAT='NSARAT'
    NSARATAD='NSARATAD'
    NSARB='NSARB'
    NSARBAD='NSARBAD'
    NSARBT='NSARBT'
    NSARBTAD='NSARBTAD'
    NSARU='NSARU'
    NSARUAD='NSARUAD'
    NT10D='NT10D'
    NT10DAD='NT10DAD'
    NT10K='NT10K'
    NT10KAD='NT10KAD'
    NT10Q='NT10Q'
    NT10QAD='NT10QAD'
    NT11K='NT11K'
    NT11KAD='NT11KAD'
    NT15D2='NT15D2'
    NT15D2AD='NT15D2AD'
    NT20F='NT20F'
    NT20FAD='NT20FAD'
    NTFNCEN='NTFNCEN'
    NTFNCSR='NTFNCSR'
    NTFNSAR='NTFNSAR'
    NTN10D='NTN10D'
    NTN10K='NTN10K'
    NTN10Q='NTN10Q'
    NTN15D2='NTN15D2'
    NTN20F='NTN20F'
    NTNCEN='NTNCEN'
    NTNCENX='NTNCENX'
    NTNCENXDA='NTNCENXDA'
    NTNCSR='NTNCSR'
    NTNCSRAD='NTNCSRAD'
    NTNMFP='NTNMFP'
    NTNMFPAD='NTNMFPAD'
    NTNMFP1='NTNMFP1'
    NTNMFP1AD='NTNMFP1AD'
    NTNMFP2='NTNMFP2'
    NTNMFP2AD='NTNMFP2AD'
    NTNSAR='NTNSAR'
    NTNSARAD='NTNSARAD'
    OIPNTC='OIPNTC'
    OIPORDR='OIPORDR'
    POS462B='POS462B'
    POS462C='POS462C'
    POS8C='POS8C'
    POSAM='POSAM'
    POSAMC='POSAMC'
    POSAMI='POSAMI'
    POSASR='POSASR'
    POSEX='POSEX'
    PRE14A='PRE14A'
    PRE14C='PRE14C'
    PREC14A='PREC14A'
    PREC14C='PREC14C'
    PREM14A='PREM14A'
    PREM14C='PREM14C'
    PREN14A='PREN14A'
    PRER14A='PRER14A'
    PRER14C='PRER14C'
    PRES14A='PRES14A'
    PRES14C='PRES14C'
    PRRN14A='PRRN14A'
    PX14A6G='PX14A6G'
    PX14A6N='PX14A6N'
    QRTLYRPT='QRTLYRPT'
    QRTLYRPTAD='QRTLYRPTAD'
    QUALIF='QUALIF'
    RW='RW'
    RWWD='RWWD'
    S1='S1'
    S1AD='S1AD'
    S11='S11'
    S11AD='S11AD'
    S11MEF='S11MEF'
    S1MEF='S1MEF'
    S2='S2'
    S2AD='S2AD'
    S20='S20'
    S20AD='S20AD'
    S2MEF='S2MEF'
    S3='S3'
    S3AD='S3AD'
    S3ASR='S3ASR'
    S3D='S3D'
    S3DPOS='S3DPOS'
    S3MEF='S3MEF'
    S4='S4'
    S4AD='S4AD'
    S4EF='S4EF'
    S4EFAD='S4EFAD'
    S4MEF='S4MEF'
    S4POS='S4POS'
    S6='S6'
    S6AD='S6AD'
    S6EL24='S6EL24'
    S6EL24AD='S6EL24AD'
    S8='S8'
    S8POS='S8POS'
    SB='SB'
    SBAD='SBAD'
    SB1='SB1'
    SB1AD='SB1AD'
    SB1MEF='SB1MEF'
    SB2='SB2'
    SB2AD='SB2AD'
    SB2MEF='SB2MEF'
    SBMEF='SBMEF'
    SBSE='SBSE'
    SBSEA='SBSEA'
    SBSEBD='SBSEBD'
    SBSEC='SBSEC'
    SBSEW='SBSEW'
    SC13D='SC13D'
    SC13DAD='SC13DAD'
    SC13E1='SC13E1'
    SC13E1AD='SC13E1AD'
    SC13E3='SC13E3'
    SC13E3AD='SC13E3AD'
    SC13E4='SC13E4'
    SC13E4AD='SC13E4AD'
    SC13E4F='SC13E4F'
    SC13E4FAD='SC13E4FAD'
    SC13G='SC13G'
    SC13GAD='SC13GAD'
    SC14D1='SC14D1'
    SC14D1AD='SC14D1AD'
    SC14D1F='SC14D1F'
    SC14D1FAD='SC14D1FAD'
    SC14D9='SC14D9'
    SC14D9AD='SC14D9AD'
    SC14D9C='SC14D9C'
    SC14D9F='SC14D9F'
    SC14D9FAD='SC14D9FAD'
    SC14F1='SC14F1'
    SC14F1AD='SC14F1AD'
    SC14N='SC14N'
    SC14NAD='SC14NAD'
    SC14NS='SC14NS'
    SC14NSAD='SC14NSAD'
    SCTOC='SCTOC'
    SCTOI='SCTOI'
    SCTOIAD='SCTOIAD'
    SCTOT='SCTOT'
    SCTOTAD='SCTOTAD'
    SD='SD'
    SDAD='SDAD'
    SDR='SDR'
    SDRAD='SDRAD'
    SDRA='SDRA'
    SDRW='SDRW'
    SF1='SF1'
    SF1AD='SF1AD'
    SF3='SF3'
    SF3AD='SF3AD'
    SF3ME='SF3ME'
    SP15D2='SP15D2'
    SP15D2AD='SP15D2AD'
    SUPPL='SUPPL'
    T3='T3'
    T3AD='T3AD'
    T6='T6'
    T6AD='T6AD'
    TA1='TA1'
    TA1AD='TA1AD'
    TA2='TA2'
    TA2AD='TA2AD'
    TAW='TAW'
    U1='U1'
    U1AD='U1AD'
    U12IA='U12IA'
    U12IAAD='U12IAAD'
    U12IB='U12IB'
    U12IBAD='U12IBAD'
    U131='U131'
    U131AD='U131AD'
    U1360='U1360'
    U1360AD='U1360AD'
    U13E1='U13E1'
    U13E1AD='U13E1AD'
    U13El='U13El'
    U13ElAD='U13ElAD'
    U33S='U33S'
    U33SAD='U33SAD'
    U3A2='U3A2'
    U3A2AD='U3A2AD'
    U3A31='U3A31'
    U3A31AD='U3A31AD'
    U57='U57'
    U57AD='U57AD'
    U5A='U5A'
    U5AAD='U5AAD'
    U5B='U5B'
    U5BAD='U5BAD'
    U5S='U5S'
    U5SAD='U5SAD'
    U6B2='U6B2'
    U6B2AD='U6B2AD'
    U7D='U7D'
    U7DAD='U7DAD'
    U9C3='U9C3'
    U9C3AD='U9C3AD'
    UNDER='UNDER'
    UNDERAD='UNDERAD'
    UPLOAD='UPLOAD'
    UR1='UR1'
    UR1AD='UR1AD'
    X17A5='X17A5'
    X17A5AD='X17A5AD'

from enum import Enum


class FilingType(Enum):
    """Available filing types to be used when creating Filing object.
    .. versionadded:: 0.1.5
    """

    FILING_1 = '1'
    FILING_1A = '1-a'
    FILING_1E = '1-e'
    FILING_1K = '1-k'
    FILING_1N = '1-n'
    FILING_1SA = '1-sa'
    FILING_1U = '1-u'
    FILING_1Z = '1-z'
    FILING_10 = '10'
    FILING_10D = '10-d'
    FILING_10K = '10-k'
    FILING_10M = '10-m'
    FILING_10Q = '10-q'
    FILING_11K = '11-k'
    FILING_12B25 = '12b-25'
    FILING_13F = '13f'
    FILING_13H = '13h'
    FILING_144 = '144'
    FILING_15 = '15'
    FILING_15F = '15f'
    FILING_17H = '17-h'
    FILING_18 = '18'
    FILING_18K = '18-k'
    FILING_19B4 = '19b-4'
    FILING_19B4E = '19b-4(e)'
    FILING_19B7 = '19b-7'
    FILING_2E = '2-e'
    FILING_20F = '20-f'
    FILING_24F2 = '24f-2'
    FILING_25 = '25'
    FILING_3 = '3'
    FILING_4 = '4'
    FILING_40F = '40-f'
    FILING_8K = '8-k'
    FILING_ABS15G = 'abs-15g'
    FILING_ABSEE = 'abs-ee'
    FILING_ADV = 'adv'
    FILING_ADVE = 'adv-e'
    FILING_ADVH = 'adv-h'
    FILING_ADVNR = 'adv-nr'
    FILING_ADVW = 'adv-w'
    FILING_ATS = 'ats'
    FILING_ATSN = 'ats-n'
    FILING_ATSR = 'ats-r'
    FILING_BD = 'bd'
    FILING_BDN = 'bd-n'
    FILING_BDW = 'bdw'
    FILING_C = 'c'
    FILING_CA1 = 'ca-1'
    FILING_CB = 'cb'
    FILING_CFPORTAL = 'cfportal'
    FILING_CUSTODY = 'custody'
    FILING_D = 'd'
    FILING_DEF_14A = 'def 14a'
    FILING_F1 = 'f-1'
    FILING_F10 = 'f-10'
    FILING_F3 = 'f-3'
    FILING_F4 = 'f-4'
    FILING_F6 = 'f-6'
    FILING_F7 = 'f-7'
    FILING_F8 = 'f-8'
    FILING_F80 = 'f-80'
    FILING_FN = 'f-n'
    FILING_FX = 'f-x'
    FILING_ID = 'id'
    FILING_MA = 'ma'
    FILING_MAI = 'ma-i'
    FILING_MANR = 'ma-nr'
    FILING_MAW = 'ma-w'
    FILING_MSD = 'msd'
    FILING_MSDW = 'msdw'
    FILING_N14 = 'n-14'
    FILING_N17D1 = 'n-17d-1'
    FILING_N17F1 = 'n-17f-1'
    FILING_N17F2 = 'n-17f-2'
    FILING_N18F1 = 'n-18f-1'
    FILING_N1A = 'n-1a'
    FILING_N2 = 'n-2'
    FILING_N23C3 = 'n-23c-3'
    FILING_N27D1 = 'n27d-1'
    FILING_N3 = 'n-3'
    FILING_N4 = 'n-4'
    FILING_N5 = 'n-5'
    FILING_N54A = 'n-54a'
    FILING_N54C = 'n-54c'
    FILING_N6 = 'n-6'
    FILING_N6EI1 = 'n-6ei-1'
    FILING_N6F = 'n-6f'
    FILING_N8A = 'n-8a'
    FILING_N8B2 = 'n-8b-2'
    FILING_N8B4 = 'n-8b-4'
    FILING_N8F = 'n-8f'
    FILING_NCEN = 'n-cen'
    FILING_NCR = 'n-cr'
    FILING_NCSR = 'n-csr'
    FILING_NCSRS = 'n-csrs'
    FILING_NLIQUID = 'n-liquid'
    FILING_NMFP = 'n-mfp'
    FILING_NPORT = 'nport'
    FILING_NPORTP = 'nport-p'
    FILING_NPORTEX = 'nport-ex'
    FILING_NPX = 'n-px'
    FILING_NQ = 'n-q'
    FILING_NRSRO = 'nrsro'
    FILING_PF = 'pf'
    FILING_PILOT = 'pilot'
    FILING_R31 = 'r31'
    FILING_S1 = 's-1'
    FILING_S11 = 's-11'
    FILING_S20 = 's-20'
    FILING_S3 = 's-3'
    FILING_S4 = 's-4'
    FILING_S6 = 's-6'
    FILING_S8 = 's-8'
    FILING_SBSE = 'sbse'
    FILING_SBSEA = 'sbse-a'
    FILING_SBSEBD = 'sbse-bd'
    FILING_SBSEC = 'sbse-c'
    FILING_SBSEW = 'sbse-w'
    FILING_SCI = 'sci'
    FILING_SD = 'sd'
    FILING_SDR = 'sdr'
    FILING_SE = 'se'
    FILING_SF1 = 'sf-1'
    FILING_SF3 = 'sf-3'
    FILING_SIP = 'sip'
    FILING_T1 = 't-1'
    FILING_T2 = 't-2'
    FILING_T3 = 't-3'
    FILING_T4 = 't-4'
    FILING_T6 = 't-6'
    FILING_TA1 = 'ta-1'
    FILING_TA2 = 'ta-2'
    FILING_TAW = 'ta-w'
    FILING_TCR = 'tcr'
    FILING_TH = 'th'
    FILING_WBAPP = 'wb-app'


from setuptools import setup, find_packages

# load the README file.
with open(file="README.md", mode="r") as readme_file:
    long_description = readme_file.read()

setup(

    # this will be my Library name.
    name='python-sec',

    # Want to make sure people know who made it.
    author='Alex Reed',

    # also an email they can use to reach out.
    author_email='coding.sigma@gmail.com',

    # I'm in alpha development still, so a compliant version number is a1.
    # read this as MAJOR VERSION 0, MINOR VERSION 1, MAINTENANCE VERSION 0
    version='0.1.1',

    # here is a simple description of the library, this will appear when someone searches for the library on https://pypi.org/search
    description='A client library for collecting and scraping SEC filings.',

    # I have a long description but that will just be my README file, note the variable up above where I read the file.
    long_description=long_description,

    # want to make sure that I specify the long description as MARKDOWN.
    long_description_content_type="text/markdown",

    # here is the URL you can find the code, this is just the GitHub URL.
    url='https://github.com/areed1192/python-sec',

    # there are some dependencies to use the library, so let's list them out.
    install_requires=[
        'requests>=2.24.0',
        'urllib3==1.25.9',
        'beautifulsoup4==4.9.1'
    ],

    # some keywords for my library.
    keywords='finance, sec, api, web scraping, financial disclosures',

    # here are the packages I want "build."
    packages=find_packages(include=['pysec']),

    # here we specify any package data.
    package_data={
        "pysec": ["parsing/*"],
    },

    # I also have some package data, like photos and JSON files, so I want to include those too.
    include_package_data=True,

    # additional classifiers that give some characteristics about the package.
    classifiers=[

        # I want people to know it's still early stages.
        'Development Status :: 3 - Alpha',

        # My Intended audience is mostly those who understand finance.
        'Intended Audience :: Financial and Insurance Industry',

        # My License is MIT.
        'License :: OSI Approved :: MIT License',

        # I wrote the client in English
        'Natural Language :: English',

        # The client should work on all OS.
        'Operating System :: OS Independent',

        # The client is intendend for PYTHON 3
        'Programming Language :: Python :: 3'
    ],

    # you will need python 3.7 to use this libary.
    python_requires='>=3.7'

)



import re
import xml.etree.ElementTree as ET
import requests

from pprint import pprint
from typing import List
from typing import Dict
from typing import Union

from html.parser import HTMLParser
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from bs4 import BeautifulSoup
from bs4 import Tag
from bs4 import NavigableString

class EDGARParser():

    def __init__(self):
        """Initalizes the `EDGARParser()` Object.

        Parsing filings, can change depending on the filing you're working with
        and whether you're grabbing the raw filing text or the directory of the filings.
        Regardless of what you're parsing, the `EDGARParser()` object will handle most of
        the finer details for you.

        In cases, where the user needs to parse RSS feeds for the company search, then the
        parser will grab all the XML content and convert it to a Python dictionary. Additionally,
        it will grab all the next pages and parse thoses if specified.
        """

        self.entries_namespace = {
            'atom': "http://www.w3.org/2005/Atom",
            'atom_with_quote':'{http://www.w3.org/2005/Atom}',
            '':''
        }

        self.retry_strategy = Retry(
            total=3,
            backoff_factor=0.2,
            status_forcelist=[429, 500, 502, 503, 504]
        )
        self.adapter = HTTPAdapter(max_retries=self.retry_strategy)

    def parse_entries(self, entries_text: str, num_of_items: int = None, start: int = None) -> List[Dict]:
        """Parses all the entries from an entry element list.

        Arguments:
        ----
        entries_text {str} -- The raw string returned from the
            response.

        Returns:
        ----
        List[Dict] -- A dictionary containing all the information from the
            original entry element.
        """        

        # Parse the text.
        root = ET.fromstring(entries_text)
        entries = []
        keep_going = True
        
        if start:
            current_count = start
        else:
            current_count = 0

        while keep_going:

            # Check for the next page Link, if there is one.
            next_page = self._check_for_next_page(root_document=root)
            
            # Grab the next page.
            if next_page and start:
                current_count = (int(next_page.split('&start=')[1]) - start)
            elif next_page:
                current_count = int(next_page.split('&start=')[1])

            # Find all the entries.
            for entry in root.findall('atom:entry', namespaces=self.entries_namespace):
                
                # Parse the individual entry.
                entry_dict = self.parse_entry_element(entry=entry)
                entries.append(entry_dict)

            # If there is a next page continue.
            if not next_page:
                keep_going = False
            else:
                root = self._grab_next_page(next_url=next_page)
                print('Grabbed Next URL: {url}'.format(url=next_page))
                
                if not root or (num_of_items and num_of_items < current_count):
                    keep_going = False

        return entries
    
    def _grab_next_page(self, next_url: str) -> ET.ElementTree:
        """Grabs the next page text content.

        Grabbing mutliple pages can be challenging because in some
        cases the SEC will kick you back if you make too many requests
        at once and don't pause enough. This method will help control that
        by defining a retry strategy and backing off for an allotted time
        in the case of a failed request.

        Arguments:
        ----
        next_url {str} -- URL redirecting to the next rounds of files.

        Returns:
        ----
        ET.ElementTree -- A parsed version of the RSS Feed.
        """

        # Create a new session.
        http = requests.Session()

        # Set the retry strategy.
        http.mount("https://", self.adapter)

        # Make the request.
        try:
            entries_response = http.get(url=next_url)
        except:
            return None

        # If it was successful, get the data.
        if entries_response.status_code == 200:
            root = ET.fromstring(entries_response.content)
            return root
        else:
            return None

    def parse_entry_element(self, entry: ET.ElementTree) -> dict:
        """Converts the XML entry element into a python dictionary.

        Arguments:
        ----
        entry {ET.ElementTree} -- An entry element, that contains filing information.

        Returns:
        ----
        dict -- A dictionary version of the entry element.
        """        

        entry_element_dict = {}
        replace_tag = self.entries_namespace['atom_with_quote']

        for entry in entry.findall("./", namespaces=self.entries_namespace):
            for element in entry.iter():
                name = element.tag.replace(replace_tag, '')
                
                if element.text :
                    name = name.replace('-','_')
                    entry_element_dict[name] = element.text.strip()

                if element.attrib:
                    for key, value in element.attrib.items():
                        key = key.replace('-','_')
                        entry_element_dict[name + "_{}".format(key)] = value

        return entry_element_dict

    def _check_for_next_page(self, root_document: ET.Element) -> Union[str, None]:
        """Checks if the RSS Feed has a next page.

        Arguments:
        ----
        root_document {ET.Element} -- The Parsed root document, which contains entry
            elements.

        Returns:
        ----
        Union[str, None] -- The URL if it was found otherwise nothing.
        """        

        next_page = root_document.findall("atom:link[@rel='next']", namespaces=self.entries_namespace)

        if next_page:
            element_attributes = next_page[0].attrib
        else:
            return None

        if 'href' in element_attributes:
            next_page_url = element_attributes['href']
        else:
            return None

        return next_page_url

    def _parse_issuer_next_button(self, button_soup: Tag) -> Union[str]:
        """Parses the next button in the issuer report.

        Args:
        ----
        button_soup (Tag): The raw HTML of the page with the button included.

        Returns:
        ----
        Union[str]: The link to the next page or nothing.
        """        
        
        # Grab the button.
        buttons: Tag = button_soup.find_all(
            name='input',
            attrs={'type': 'button'}
        )

        # Loop through each of the buttons.
        for button in buttons:

            # If there is a next button, grab the link.
            if 'Next' in button['value']:
                next_page = button['onclick']

                # Build the URL
                next_page_link = next_page.replace("parent.location='","https://www.sec.gov").replace("'","")

                return next_page_link

    def parse_issuer_table(self, entries_text: str, num_of_items: int = None) -> List[Dict]:
        """Parses the Issuer tables found from a query to owner distribtuion page.

        Arguments:
        ----
        entries_text (str): The raw HTML content to be parsed.

        num_of_items (int, optional): The number of items to return from the query. Defaults to None.

        Returns:
        ----
        List[Dict]: A list of dictionaries where each dictionary contains the `ownership_report`
            and the `ownership_transaction_report`.
        """        

        master_list = []
        ownership_report_for_issuers = []

        soup = BeautifulSoup(entries_text, 'html.parser')
        next_page_link = self._parse_issuer_next_button(button_soup=soup)

        while soup is not None:

            # table_rows = soup.find_all(name='tr')
            table = soup.find_all(name='table')

            issuers_table: Tag = table[4]
            issuers_transaction_report_table: Tag = soup.find_all(
                name='table',
                attrs={
                    'id':'transaction-report'
                }
            )
            issuers_table_rows = issuers_table.find_all(name='tr')
            
            for row in issuers_table_rows:

                issuer_dict = {}

                row: Tag = row
                elements = row.text.split('\n')
                links = row.find_all('a', href=True)

                issuer_dict = {
                    'issuer': elements[0],
                    'filings': elements[1],
                    'transaction_date': elements[2],
                    'type_of_owner': elements[3]
                }

                ownership_report_for_issuers.append(issuer_dict)

                for link in links:
                    new_link = 'https://www.sec.gov' + link['href']

                    if 'own-disp' in new_link:
                        issuer_dict['ownership_link'] = new_link
                    elif 'browse-edgar' in new_link:
                        issuer_dict['browse_company_link'] = new_link       
            
            master_dict = {}
            master_dict['ownership_report'] = ownership_report_for_issuers
            master_dict['ownership_transaction_report'] = self.parse_transaction_report(
                table=issuers_transaction_report_table[0]
            )

            master_list.append(master_dict)

            print("Pulling URL: {url}".format(url=next_page_link))
            
            if next_page_link:
                entries_text = requests.get(next_page_link).content
                soup = BeautifulSoup(entries_text, 'html.parser')
                next_page_link = self._parse_issuer_next_button(button_soup=soup)
            else:
                soup = None

        return master_list

    def parse_transaction_report(self, table: Tag) -> List[Dict]:
        """Parse the transaction table report.

        Arguments:
        ----
        table (Tag): The raw HTML table.

        Returns:
        ----
        List[Dict]: A list of ownership transaction reports.
        """        

        master_list = []
        
        all_rows = table.find_all('tr')
        first_row = all_rows[0]
        all_other_rows = all_rows[1:]

        headers = [header.replace(' ', '_').lower() for header in first_row.strings if header != '\n']

        headers.insert(5, 'link')

        for row in all_other_rows:

            link = row.find_all('a')[0]
            href = 'https://www.sec.gov' + link['href']

            values = [value.strip() for value in row.strings if value != '\n']
            values.insert(5, href)

            master_list.append(dict(zip(headers, values)))
            # print([header.strip() for header in row.strings if header != '\n'])

        return master_list

    def _check_center_tag(self, product_table_soup: Tag) -> Union[List[str]]:
        """Grabs all the links that are in the Center tag of the Product Page.

        Arguments:
        ----
        product_table_soup (Tag): The product page HTML that has been parsed.

        Returns:
        ----
        Union[List[str]]: A list of URL links to the other pages.
        """

        links = []

        # Find the <center> tag.
        center_tag: Tag = product_table_soup.find(name='center')

        # Grab the <A> tags.
        center_tag_hrefs = center_tag.find_all('a', href=True)

        # We need to clean up the links, cause they're broken.
        for href in center_tag_hrefs:

            # Split by the ampersand.
            split_href: list = href['href'].split('&')

            # Remove the CIK parameters.
            split_href.pop(1)

            # and add a clean one.
            split_href.insert(1, 'CIK=')

            # Create a new URL.
            links.append('https://www.sec.gov' + '&'.join(split_href))

        # Don't care about the last one, just a duplicate.
        return links[:-1]
    
    def parse_variable_products_company_table(self, product_table_page: str) -> List[Dict]:
        """Parses the Variable Product page of all the different products and companies.

        Arguments:
        ----
        product_table_page (str): The raw HTML of the Product query result page.

        Returns:
        ----
        List[Dict]: A list of variable products.
        """
        # Parse the Page.
        product_page_soup = BeautifulSoup(product_table_page, 'html.parser')

        # Check for the other links.
        href_links = self._check_center_tag(product_table_soup=product_page_soup)

        # Parse the table.
        product_list_all = self._parse_variable_product_page(product_page_soup=product_page_soup)

        # Loop through all the pages, and grab those entries.
        for link in href_links:

            product_page_soup = BeautifulSoup(requests.get(url=link).text, 'html.parser')
            product_list = self._parse_variable_product_page(product_page_soup=product_page_soup)
            product_list_all = product_list_all + product_list

            print("Pulling URL: {url}".format(url=link))
            print("Total Entries Scraped: {quant}".format(quant=len(product_list_all)))
        
        return product_list_all

    def _parse_variable_product_page(self, product_page_soup: Tag) -> List[Dict]:
        """This parses the actual table. It will grab the table with the entires and parse each row.

        Arguments:
        ----
        product_page_soup (Tag): The parsed page with product tables.

        Returns:
        ----
        List[Dict]: A list of variable products.
        """        

        # Grab all the Summary Tables.
        summary_table = product_page_soup.find_all(name='table', attrs={'summary':'.'})

        master_list = []
        
        # Loop through each table.
        for table in summary_table:

            table: Tag = table

            # Define a matching table.
            criteria_1 = len(table.attrs) == 1
            criteria_2 = table.attrs['summary'] == '.'
            criteria_3 = len(table.find_all('input')) == 0

            if criteria_1 and criteria_2 and criteria_3:

                # Grab all the Rows.
                table_rows: List[Tag] = table.find_all('tr', attrs={'valign':'top'})
                
                # Loop through each Row.
                for row in table_rows:
                    
                    # Grab Row links
                    row_links = [row_link['href'] for row_link in row.find_all('a', href=True)]
                    
                    # Grab the Strings, with text and filter our line breaks.
                    values = [string for string in row.strings if string != '\n']

                    if values:

                        # Create the dictionary & store the values.
                        row_dict = {}
                        product_id: str = values[0]
                        product_name: str = values[1]

                        row_dict['id'] = product_id
                        row_dict['name'] = product_name
                        
                        # Define the product ID based on the first Character.
                        if product_id.startswith('S'):
                            row_dict['id_type'] = 'Series'
                        elif product_id.startswith('C'):
                            row_dict['id_type'] = 'Contract'
                        else:
                            row_dict['id_type'] = 'CIK'
                        
                        # Set the Ticker symbol.
                        try:
                            row_dict['ticker_symbol'] = values[2]
                        except:
                            row_dict['ticker_symbol'] = "null"

                        for link in row_links:

                            if '&scd' in link:
                                row_dict['series_link'] = 'https://www.sec.gov' + link
                            elif 'getcompany' in link and row_dict['id_type'] == 'Contract':
                                row_dict['contract_id_link'] = 'https://www.sec.gov' + link
                            elif 'getcompany' in link and row_dict['id_type'] == 'Series':
                                row_dict['series_id_link'] = 'https://www.sec.gov' + link
                            else:
                                row_dict['cik_id_link'] = 'https://www.sec.gov' + link

                        master_list.append(row_dict)

        return master_list
    
    def parse_current_event_table(self, current_event_page: str) -> List[Dict]:
        """Parses the Current Event page of all the forms.

        Arguments:
        ----
        current_event_page (str): The raw HTML of the current event query page.

        Returns:
        ----
        List[Dict]: A list of SEC filings.
        """

        master_list = []

        # Parse the Page.
        current_event_soup = BeautifulSoup(current_event_page, 'html.parser')

        # Grab the <Pre> tag.
        current_event_pre: Tag = current_event_soup.find('pre')

        # In this case, split along Line breaks.
        all_rows = current_event_pre.text.splitlines()

        # Define the headers
        keys = ['date_filed', 'form', 'cik', 'company_name']

        # Clean up the header to get the first row.
        header_row = all_rows[0].replace('Date Filed   Form        CIK Code     Company Name','')
        new_string = " ".join(header_row.split()).split(' ', 3)

        # add to the list.
        master_list.append(dict(zip(keys, new_string)))

        # Clean up the rest of the rows.
        for row in all_rows[1:]:
            new_string = " ".join(row.split()).split(' ', 3)
            master_list.append(dict(zip(keys, new_string)))

        return master_list

    def parse_loc_elements(self, response_text: str) -> List[dict]:
        """Parses the `Loc` elements found in the Taxonomies XML content.

        Arguments:
        ----
        response_text (str): The raw XML of the Taxonomy query.

        Returns:
        ----
        List[Dict]: A list of taxonomy objects..
        """

        # Parse the text.
        root = ET.fromstring(response_text)
        entries = []

        # Grab all the location elements.
        for location in root.findall('Loc'):
            location_dict = {element.tag: element.text for element in location.iter()}
            del location_dict['Loc']
            entries.append(location_dict)

        return entries
